/*
* SACD Decoder plugin
* Copyright (c) 2011 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _ASIO_OUTPUT_H_INCLUDED
#define _ASIO_OUTPUT_H_INCLUDED

#include <stdint.h>
#include <foobar2000.h>
#include <asiosys.h>
#include <asio.h>
#include <iasiodrv.h>

#define ASIO_MAXCHANNELS 16

#define DSDxFs64  (44100 * 64)
#define DSDxFs128 (44100 * 128)
#define DSDxFs1   (44100 * 1)
#define DSDxFs2   (44100 * 2)
#define DSDxFs4   (44100 * 4)
#define DSDxFs8   (44100 * 8)

typedef HRESULT CoCreateInstance_t(REFCLSID rclsid, LPUNKNOWN lpUnkOuter, DWORD dwClsContext, REFIID riid, LPVOID* ppv);

class PEImage {
	PBYTE                    p_mod;
	PIMAGE_DOS_HEADER        p_idh;
	PIMAGE_FILE_HEADER       p_ifh;
	PIMAGE_OPTIONAL_HEADER   p_ioh;
	PIMAGE_SECTION_HEADER    p_ish;
	PIMAGE_IMPORT_DESCRIPTOR p_iid;
public:
	bool map(HMODULE hModule);
	bool install_thunk(LPSTR lpszLibName, FARPROC fpOldProc, FARPROC fpNewProc);
};

enum asio_mode_t {ASIO_MODE_PCM, ASIO_MODE_DSD};

class CoASIOProxy : public IASIO {
	friend class asio_output_t;
	static ASIOCallbacks   dsd_callbacks;
	static ASIOCallbacks   pcm_callbacks;
	static ASIOCallbacks   ext_callbacks;
	static ASIOChannelInfo channels[ASIO_MAXCHANNELS];
	static ASIOBufferInfo  dsd_buffers[ASIO_MAXCHANNELS];
	static ASIOBufferInfo  pcm_buffers[ASIO_MAXCHANNELS];
	static asio_mode_t     mode;
	static CLSID           clsid;
	static IASIO*          pIASIO;
	static long            numChannels;
	static long            dsd_bufferSize;
	static long            pcm_bufferSize;
	static long            dsd_sampleSize[ASIO_MAXCHANNELS];
	static long            pcm_sampleSize[ASIO_MAXCHANNELS];
	static double          dsd_sampleRate;
	static double          pcm_sampleRate;
	static asio_mode_t     requested_mode;
	static double          requested_dsd_sampleRate;
	static double          requested_pcm_sampleRate;
public:
	CoASIOProxy();
	~CoASIOProxy();
	void request_mode(asio_mode_t mode) {
		requested_mode = mode;
	}
	void request_dsd_samplerate(int samplerate) {
		requested_dsd_sampleRate = samplerate;
	}
	void request_pcm_samplerate(int samplerate) {
		requested_pcm_sampleRate = samplerate;
	}
	asio_mode_t get_mode() {
		return mode;
	}
	CLSID& get_CLSID() {
		return clsid;
	}
	void set_CLSID(const CLSID& clsid) {
		this->clsid = clsid;
	}
	IASIO* get_IASIO() {
		return pIASIO;
	}
	void set_IASIO(IASIO* pIASIO) {
		this->pIASIO = pIASIO;
	}
	void register_callbacks(ASIOCallbacks* callbacks) {
		memcpy(&ext_callbacks, callbacks, sizeof(ASIOCallbacks));
	}
	void unregister_callbacks() {
		memset(&ext_callbacks, 0, sizeof(ASIOCallbacks));
	}

	// IUnknown
	STDMETHODIMP QueryInterface(REFIID riid, LPVOID* ppv);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();

	// IASIO
	ASIOBool init(void* sysRef);
	void getDriverName(char* name);
	long getDriverVersion();
	void getErrorMessage(char* string);
	ASIOError start();
	ASIOError stop();
	ASIOError getChannels(long* numInputChannels, long* numOutputChannels);
	ASIOError getLatencies(long* inputLatency, long* outputLatency);
	ASIOError getBufferSize(long* minSize, long* maxSize, long* preferredSize, long* granularity);
	ASIOError canSampleRate(ASIOSampleRate sampleRate);
	ASIOError getSampleRate(ASIOSampleRate* sampleRate);
	ASIOError setSampleRate(ASIOSampleRate sampleRate);
	ASIOError getClockSources(ASIOClockSource* clocks, long* numSources);
	ASIOError setClockSource(long index);
	ASIOError getSamplePosition(ASIOSamples* sPos, ASIOTimeStamp* tStamp);
	ASIOError getChannelInfo(ASIOChannelInfo* info);
	ASIOError createBuffers(ASIOBufferInfo* bufferInfos, long numChannels, long bufferSize, ASIOCallbacks* callbacks);
	ASIOError disposeBuffers();
	ASIOError controlPanel();
	ASIOError future(long selector, void* opt);
	ASIOError outputReady();
private:
	static ASIOError set_mode(asio_mode_t mode);
	static long getSampleSizeInBytes(ASIOSampleType type);
	static long getDsdBufSize(long pcm_bufSize);
	static long getPcmBufSize(long dsd_bufSize);

	// ASIO Callbacks
	static void bufferSwitch(long doubleBufferIndex, ASIOBool directProcess);
	static void sampleRateDidChange(ASIOSampleRate sRate);
	static long asioMessage(long selector, long value, void* message, double* opt);
	static ASIOTime* bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, ASIOBool directProcess);
};

enum asio_state_t {ASIO_INITIAL = 0, ASIO_LOADED = 1 << 0, ASIO_INITIALIZED = 1 << 1, ASIO_PREPARED = 1 << 2, ASIO_RUNNING = 1 << 3};

class asio_block_t {
public:
	uint8_t*      data;
	long          size;
	long          start;
	asio_block_t* next;
	asio_block_t() {
		data = NULL;
		size = 0;
		start = 0;
		next = NULL;
	}
};

class asio_output_t {
#pragma pack(4)
	volatile long       cached_blocks;
	volatile long       flush_blocks;
#pragma pack()
	ASIOCallbacks       dsd_callbacks;
	asio_block_t*       p_get;
	asio_block_t*       p_put;
	HMODULE             hAsioOutput;
	CoCreateInstance_t* fpCoCreateInstance;
public:
	asio_output_t();
	~asio_output_t();
	bool hook(bool install);
	bool create(int channels, int samplerate);
	bool dispose();
	void get_data(long doubleBufferIndex);
	bool set_data(uint8_t* dsd_data, size_t dsd_size);
	void playback_start();
	void playback_stop();
	void playback_seek();
	void playback_pause(bool playback_state);
private:
	inline long get_cached_blocks() {
		return cached_blocks;
	}
	inline long get_flush_blocks() {
		return flush_blocks;
	}
	inline void set_cached_blocks(long cached_blocks) {
		InterlockedExchange(&this->cached_blocks, cached_blocks);
	}
	inline void set_flush_blocks(long flush_blocks) {
		InterlockedExchange(&this->flush_blocks, flush_blocks);
	}
	inline void increment_cached_blocks() {
		InterlockedIncrement(&cached_blocks);
	}
	inline void decrement_cached_blocks() {
		InterlockedDecrement(&cached_blocks);
	}
	inline void decrement_flush_blocks() {
		InterlockedDecrement(&flush_blocks);
	}
	void flush_data();
	void request_flush() {
		set_flush_blocks(cached_blocks);
	};
};

extern asio_output_t g_asio_output;
extern CoASIOProxy   g_proxy;

#endif
